use std::ops::Neg;

/// Marker trait, because certain operations only make sense on floating point types.
pub trait Float: Neg<Output=Self> {
    /// Calculate the absolute value. For primitive floats it just calls their `abs` function.
    fn fabs(self) -> Self;

    /// Calculate the square root. For primitive floats it just calls their `sqrt` function.
    fn fsqrt(self) -> Self;

    /// Calculate the sine.
    fn fsin(self) -> Self;

    /// Calculate the cosine.
    fn fcos(self) -> Self;

    /// Calculate the hyperbolic sine.
    fn fsinh(self) -> Self;

    /// Calculate the hyperbolic cosine.
    fn fcosh(self) -> Self;
}

impl Float for f32 {
    fn fabs(self) -> Self {
        self.abs()
    }

    fn fsqrt(self) -> Self {
        self.sqrt()
    }

    fn fsin(self) -> Self {
        self.sin()
    }

    fn fcos(self) -> Self {
        self.cos()
    }

    fn fsinh(self) -> Self {
        self.sinh()
    }

    fn fcosh(self) -> Self {
        self.cosh()
    }
}

impl Float for f64 {
    fn fabs(self) -> Self {
        self.abs()
    }

    fn fsqrt(self) -> Self {
        self.sqrt()
    }

    fn fsin(self) -> Self {
        self.sin()
    }

    fn fcos(self) -> Self {
        self.cos()
    }

    fn fsinh(self) -> Self {
        self.sinh()
    }

    fn fcosh(self) -> Self {
        self.cosh()
    }
}

#[cfg(test)]
mod tests {
    use crate::Float;

    fn is_float<T: Float>() -> bool {
        true
    }

    #[test]
    fn f32_is_a_float() {
        assert!(is_float::<f64>());
    }

    #[test]
    fn f64_is_a_float() {
        assert!(is_float::<f64>());
    }
}
